#-*- coding: utf8 -*
#
# Max E. Kuznecov ~syhpoon <syhpoon@syhpoon.name> 2008
#
# This file is part of XYZCommander.
# XYZCommander is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# XYZCommander is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser Public License for more details.
# You should have received a copy of the GNU Lesser Public License
# along with XYZCommander. If not, see <http://www.gnu.org/licenses/>.

import os
import re

import libxyz.parser as parser

from libxyz.exceptions import ParseError
from libxyz.exceptions import SkinError
from libxyz.exceptions import XYZValueError
from libxyz.exceptions import FSRuleError
from libxyz.core import FSRule

import libxyz.ui as uilib

class Skin(object):
    """
    Skin object. Provides simple interface to defined skin rulesets.
    """

    def __init__(self, path):
        """
        @param path: Path to skin file
        """

        if not os.access(path, os.R_OK):
            raise SkinError(_(u"Unable to open skin file for reading"))
        else:
            self.path = path

        self._data = {}

        self.screen = None

        # Default fallback palette
        self._default = uilib.colors.Palette(u"default",
                        uilib.colors.Foreground(u"DEFAULT"),
                        uilib.colors.Background(u"DEFAULT"),
                        uilib.colors.Monochrome(u"DEFAULT"))

        # 1. Parse
        self._data = self._parse()

        # 2. Order parsed data
        self._check()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def __str__(self):
        return u"<Skin object: %s>" % str(self.path)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def __repr__(self):
        return self.__str__()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def __getitem__(self, key):
        return self._data[key]

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def set_screen(self, screen):
        self.screen = screen

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _parse(self):
        def palette_validator(block, var, val):
            """
            Make L{libxyz.ui.colors.Palette} object of palette definition
            """

            _p = self._default.copy()

            if isinstance(val, basestring):
                _val = (val,)
            else:
                _val = [x.strip() for x in val]

            _p.fg = uilib.colors.Foreground(_val[0])

            if len(_val) > 1:
                _p.bg = uilib.colors.Background(_val[1])
            if len(_val) > 2:
                _p.ma = tuple([uilib.colors.Monochrome(x) for x in _val[2:]])

            _p.name = self._make_name(block, var)

            return _p

        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        def trans_cr(rule):
            """
            Transform string rules to FSRule objects
            """

            try:
                return FSRule(rule)
            except (ParseError, FSRuleError), e:
                raise XYZValueError(e)

        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        # Prepare parsers

        _fs_rules_opt = {u"count": 1,
                         u"value_validator": palette_validator,
                         u"varre": re.compile(r".+"),
                         u"var_transform": trans_cr,
                         }

        _fs_rules_p = parser.BlockParser(_fs_rules_opt)

        _ui_opt = {u"count": 1,
                   u"value_validator": palette_validator,
                   }
        _ui_p = parser.BlockParser(_ui_opt)

        _plugin_opt = {u"count": 1,
                       u"value_validator": palette_validator,
                       }
        _plugin_p = parser.BlockParser(_plugin_opt)

        _flat_opt = {u"count": 1}
        _flat_p = parser.FlatParser(_flat_opt)

        _parsers = {
                    u"fs.rules": _fs_rules_p,
                    re.compile(r"ui\.(\w)+"): _ui_p,
                    re.compile(r"plugin\.([\w_-])+"): _plugin_p,
                    (u"AUTHOR", u"VERSION", u"DESCRIPTION"): _flat_p,
                    }

        _multi_opt = {u"tokens": (":",)}
        _multi_p = parser.MultiParser(_parsers, _multi_opt)

        _skinfile = open(self.path, "r")

        try:
            _data = _multi_p.parse(_skinfile)
        except ParseError, e:
            raise SkinError(_(u"Error parsing skin file: %s" % str(e)))
        finally:
            _skinfile.close()

        return _data

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _make_name(self, block, resource):
        return "%(resource)s@%(block)s" % locals()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _check(self):
        """
        Check and variables
        """

        for _required in (u"AUTHOR", u"DESCRIPTION", u"VERSION"):
            if _required not in self._data:
                raise SkinError(_(u"Missing required variable: %s"%_required))

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def get_palette_list(self):
        """
        Return list of defined palettes.
        It is usually passed to register_palette() function
        """

        _list = [self._default.get_palette()]

        for _name, _pdata in self._data.iteritems():
            if not isinstance(_pdata, parser.ParsedData):
                continue

            for _var, _val in _pdata.iteritems():
                if isinstance(_val, uilib.colors.Palette):
                    _list.append(_val.get_palette())

        return _list

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def attr(self, resolution, name, default=True):
        """
        Search for first matching attribute <name> according to resolution
        @param resolution: Sequence of ruleset names
        @param name: Attribute name
        @param default: If True, return default palette in case attr
                        is not found, otherwise return None
        @return: Registered palette name
        """

        return self.palette(resolution, name, default).name

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def palette(self, resolution, name, default=True):
        """
        Search for first matching palette <name> according to resolution
        """

        for _w in resolution:
            # Normalize name
            if not _w.startswith(u"ui."):
                _w = u"ui.%s" % _w

            try:
                return self._data[_w][name]
            except KeyError:
                pass

        if default:
            return self._default
        else:
            return None

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def get_palette(self, block, name):
        try:
            return self._data[block][name].name
        except KeyError:
            return None
